'use strict';

/**
 * SGF parser: from string to internal structure
 * node

Properties by type

ID   Description     property type    property value
---- --------------- ---------------  --------------------------------------
B    Black           move             move
BL   Black time left move             real
BM   Bad move        move             double
DO   Doubtful        move             none
IT   Interesting     move             none
KO   Ko              move             none
MN   set MoveNumber  move             number
OB   OtStones Black  move             number
OW   OtStones White  move             number
TE   Tesuji          move             double
W    White           move             move
WL   White time left move             real

AB   Add Black       setup            list of stone
AE   Add Empty       setup            list of point
AW   Add White       setup            list of stone
PL   Player to play  setup            color

*AR  Arrow           -                list of composed point ':' point
C    Comment         -                text
CR   Circle          -                list of point
*DD  Dim points      - (inherit)      elist of point
DM   Even position   -                double
!FG  Figure          -                none | composed number ":" simpletext
GB   Good for Black  -                double
GW   Good for White  -                double
HO   Hotspot         -                double
!LB  Label           -                list of composed point ':' simpletext
*LN  Line            -                list of composed point ':' point
MA   Mark            -                list of point
N    Nodename        -                simpletext
*PM  Print move mode - (inherit)      number
SL   Selected        -                list of point
*SQ  Square          -                list of point
TR   Triangle        -                list of point
UC   Unclear pos     -                double
V    Value           -                real
*VW  View            - (inherit)      elist of point

*AP  Application     root	      composed simpletext ':' number
*CA  Charset         root	      simpletext
FF   Fileformat      root	      number (range: 1-4)
GM   Game            root	      number (range: 1-5,7-17)
*ST  Style           root	      number (range: 0-3)
!SZ  Size            root	      (number | composed number ':' number)

AN   Annotation      game-info        simpletext
BR   Black rank      game-info        simpletext
BT   Black team      game-info        simpletext
CP   Copyright       game-info        simpletext
!DT  Date            game-info        simpletext
EV   Event           game-info        simpletext
GC   Game comment    game-info        text
GN   Game name       game-info        simpletext
ON   Opening         game-info        simpletext
*OT  Overtime        game-info        simpletext
PB   Player Black    game-info        simpletext
PC   Place           game-info        simpletext
PW   Player White    game-info        simpletext
!RE  Result          game-info        simpletext
RO   Round           game-info        simpletext
!RU  Rules           game-info        simpletext
SO   Source          game-info        simpletext
TM   Timelimit       game-info        real
US   User            game-info        simpletext
WR   White rank      game-info        simpletext
WT   White team      game-info        simpletext

Go (GM[1]) specific properties
ID   Description     property type    property value
---- --------------- ---------------  --------------------------------------
TB   Territory Black -                elist of point
TW   Territory White -                elist of point

HA   Handicap        game-info        number
KM   Komi            game-info        real


FF[4] property index

ID   Description     property type    property value
---- --------------- ---------------  --------------------------------------
AB   Add Black       setup            list of stone
AE   Add Empty       setup            list of point
AN   Annotation      game-info        simpletext
*AP  Application     root	      composed simpletext ':' simpletext
*AR  Arrow           -                list of composed point ':' point
*AS  Who adds stones - (LOA)          simpletext
AW   Add White       setup            list of stone
B    Black           move             move
BL   Black time left move             real
BM   Bad move        move             double
BR   Black rank      game-info        simpletext
BT   Black team      game-info        simpletext
C    Comment         -                text
*CA  Charset         root	      simpletext
CP   Copyright       game-info        simpletext
CR   Circle          -                list of point
*DD  Dim points      - (inherit)      elist of point
DM   Even position   -                double
DO   Doubtful        move             none
!DT  Date            game-info        simpletext
EV   Event           game-info        simpletext
FF   Fileformat      root	      number (range: 1-4)
!FG  Figure          -                none | composed number ":" simpletext
GB   Good for Black  -                double
GC   Game comment    game-info        text
GM   Game            root	      number (range: 1-5,7-16)
GN   Game name       game-info        simpletext
GW   Good for White  -                double
HA   Handicap        game-info (Go)   number
HO   Hotspot         -                double
*IP  Initial pos.    game-info (LOA)  simpletext
IT   Interesting     move             none
*IY  Invert Y-axis   game-info (LOA)  simpletext
KM   Komi            game-info (Go)   real
KO   Ko              move             none
!LB  Label           -                list of composed point ':' simpletext
*LN  Line            -                list of composed point ':' point
MA   Mark            -                list of point
MN   set move number move             number
N    Nodename        -                simpletext
OB   OtStones Black  move             number
ON   Opening         game-info        simpletext
*OT  Overtime        game-info        simpletext
OW   OtStones White  move             number
PB   Player Black    game-info        simpletext
PC   Place           game-info        simpletext
PL   Player to play  setup            color
*PM  Print move mode - (inherit)      number
PW   Player White    game-info        simpletext
!RE  Result          game-info        simpletext
RO   Round           game-info        simpletext
!RU  Rules           game-info        simpletext
*SE  Markup          - (LOA)          point
SL   Selected        -                list of point
SO   Source          game-info        simpletext
*SQ  Square          -                list of point
*ST  Style           root	      number (range: 0-3)
*SU  Setup type      game-info (LOA)  simpletext
!SZ  Size            root	      (number | composed number ':' number)
TB   Territory Black - (Go)           elist of point
TE   Tesuji          move             double
TM   Timelimit       game-info        real
TR   Triangle        -                list of point
TW   Territory White - (Go)           elist of point
UC   Unclear pos     -                double
US   User            game-info        simpletext
V    Value           -                real
*VW  View            - (inherit)      elist of point
W    White           move             move
WL   White time left move             real
WR   White rank      game-info        simpletext
WT   White team      game-info        simpletext
 */

const _DEBUG_ = false;

const POINT_LIST = "pointlist"

const PROPS={
"AB": ["Add_Black","pointlist","Add black setup or handicap stones, e.g., AB[dp][pd]"],
"AE": ["Add_Empty","pointlist","Make the indicated fields empty"],
"AN": ["Annotator","simpletext","Author of the game analysis"],
"AP": ["Application","simpletext","SGF editor used"],
"AW": ["Add_White","pointlist","Add white stones"],
"B": ["Black_Move","move","Make a black move"],
"BC": ["Black_Country","simpletext","Country of the black player, e.g. CN, JP, KR, TW"],
"BL": ["Black_Time_Left","time","Time left after this move"],
"BM": ["Bad_Move" ,"double",""],
"BR": ["Black_Rank","simpletext","Rank of black player, e.g. Honorary Meijin, 9p, 6d ama, 15k"],
"BT": ["Black_Team","simpletext","Team of the black player"],
"C": ["Comment","text","Any comment"],
"CA": ["Character_Set","simpletext","Character set of the text fields (e.g., UTF-8, GB2312, SJIS, ISO-2022-KR, ISO-8859-15)"],
"CP": ["Copyright","simpletext","Copyright holder"],
"CR": ["Circle","pointlist","Mark the indicated stones with a circle"],
"DD": ["Dim","pointlist","Dim (grey out) the given points. DD[] clears any setting, i.e. it undims everything."],
"DO": ["Doubtful","double",""],
"DT": ["Date","simpletext","Date, given in Western style, preferably in ISO standard form"],
"EV": ["Event","simpletext","Tournament or other event"],
"EX": ["Expected_Move","double",""],
"FF": ["File_Format","number","SGF file format, e.g., FF[1], FF[3], FF[4]"],
"FG": ["Figure","simpletext","Division for printing"],
"GB": ["Good_for_Black","double",""],
"GC": ["Game_Comment","text","Any comment. E.g. GC[This is Cho Chikun's 1000th win as a professional], GC[Translated by Megumi Kawai], GC[Wang Lei made an illegal ko capture on move 223 in what should have been a void game due to triple ko. He was ruled to have passed on that move and the position was restored with Zhou to move]"],
"GM": ["Game","number","GM[1]: Go. Higher numbers denote other games, see GM"],
"GN": ["Game_Name","simpletext","Filename, number or identifier associated with the game"],
"GW": ["Good_for_White","double",""],
"HA": ["Handicap","number","Number of handicap stones"],
"IT": ["Interesting","double",""],
"HO": ["Hotspot","double",""],
"JD": ["Japanese_Date","simpletext","Date, given in Japanese Era, Year, Month, Day format"],
"KM": ["Komi","double","Possibly negative number of points paid by Black to White"],
"KO": ["Ko","double","Ko"],
"L": ["Letter","pointlist","Letter"],
"LB": ["Label","labellist","E.g. LB[], LB[kg:a], LB[fm:F][fj:E][fk:D][gj:C][gk:B][hi:A]"],
"LC": ["Byo_Yomi_Number","number","Number of byo-yomi periods"],
"LN": ["Line","pointlist","Line"],
"LT": ["Byo_Yomi_Length","time","Length of byo-yomi periods"],
"MA": ["Mark","pointlist","Mark the indicated stones with an 'X'"],
"MN": ["Move_Number","number","In the rootnode: the number of moves of the game. Elsewhere: the request to label the move with the given number"],
"N": ["Node_Name","simpletext","E.g. N[Diagram 1], N[Sacrifice]"],
"OB": ["Overtime_Black","number","Number of black moves still to make in this byo-yomi period"],
"OH": ["Old_Handicap","simpletext","E.g. OH[(B)BW] for sen-ai-sen, a handicap where the weaker player takes Black two out of every three games, and takes Black in the present game"],
"OM": ["Moves_Per_Overtime","simpletext","Moves_Per_Overtime"],
"ON": ["Opening_Name","double","Opening Name"],
"OP": ["Overtime_Period","simpletext","Overtime_Period"], 
"OT": ["Overtime","simpletext","Description of the type of byo-yomi used"],
"OV": ["Operator_Overhead","simpletext","Operator_Overhead"],
"OW": ["Overtime_White","number","Number of white moves still to make in this byo-yomi period"],
"PB": ["Player_Black","simpletext","Name of black player"],
"PC": ["Place","simpletext","Location of the game"],
"PL": ["Player_to_Play","color","Player to go first (B or W)"],
"PM": ["Print_Move_Mode","number","How to print"],
"PW": ["Player_White","simpletext","Name of white player"],
"RE": ["Result","simpletext","E.g. Jigo, B+3.5, W+T, Void. See RE"],
"RO": ["Round","simpletext","Tournament stage or round"],
"RU": ["Rules","simpletext","Rule set in use"],
"SC": ["Secure_Stones","pointlist","Secure Stones"],
"SL": ["Selected_Points","pointlist","Select a number of points in order to refer to them in a comment"],
"SO": ["Source","simpletext","Game source. E.g. SO[weiqi.com.tom], SO[Kido yearbook 1969]"],
"SQ": ["Square","pointlist","Mark the indicated stones with a square"],
"ST": ["Variation_Style","number","Defines how variations should be shown 0-3"],
"SZ": ["Size","numberpair","E.g. SZ[19], or SZ[2:9] for a 2x9 board (with 2 columns and 9 rows)"],
"TB": ["Territory_Black","pointlist","Black territory"],
"TC": ["Territory_Count","pointlist","Territory Count"],
"TE": ["Tesuji","double","Tesuji"],
"TM": ["Time","time","Total time for each player"],
"TR": ["Triangle","pointlist","Mark the indicated stones with a triangle"],
"TW": ["Territory_White","pointlist","White territory"],
"UC": ["Unclear_Position","double","Unclear Position"], 
"US": ["User","simpletext","User who entered the game"],
"V": ["Value","double","Value"],
"VW": ["View","pointlist","Part of the board to be shown. A subsequent VW[] removes this restriction"],
"W": ["White move","move","Make a white move"],
"WC": ["White_Country","simpletext","Country of the white player"],
"WL": ["White_Time_Left","time","Time left after this move"],
"WR": ["White_Rank","simpletext","Rank of white player"],
"WT": ["White_Team","simpletext","Team of the white player"]
}

// eslint-disable-next-line no-unused-vars
const ROOT_PROPS = ["AP","AN","BR","BT","CP","DT","EV","CA","FF","GM","ST","SZ","GN","GC","ON","OT","PB","PC","PW","RE",
                    "RO","RU","SO","TM","US","WR","WT","BC","WC","LC","LT","KM","HA","SZ","TB","TW","MN","GC","GN"]

// eslint-disable-next-line no-unused-vars
const IN_GAME_PROPS = [
"B","W","C",
"BL","WL","OB","OW",
"SL","CR","SQ","TR","MA","LB","N","MN",
"VW",
"BM","DO","IT","KO","TE","GB","GW","HO","ON","UC","V"
]

// eslint-disable-next-line no-unused-vars
const SETUP_PROPS = ["AE","AB","AW","PL"]
// eslint-disable-next-line no-unused-vars
const MOVE_PROPS = ["B","W","KO","MN"]
// eslint-disable-next-line no-unused-vars
const ANNOTATION_PROPS = ["C","DM","GB","GW","HO","N","UC","V", "BM","DO","IT","TE","TB","TW"]
// eslint-disable-next-line no-unused-vars
const MARKING_PROPS = ["N","AR","DD","LB","LN", "CR","MA","SL","SQ","TR"]
// eslint-disable-next-line no-unused-vars
const TIMING_PROPS = ["BL","OB","OW","WL"]


// eslint-disable-next-line no-unused-vars
var SGFMoveProps = {
  'B': 'blackMove',
  'W': 'whiteMove',
  'KO': 'KO',
  'MN': 'nodeNumber',
  'LB': 'label',
  'N': 'nodeName',
  
  'C': 'comment',
  'TW': 'WHITE_TERRITORY',
  'TB': 'BLACK_TERRITORY',

  'CR': 'CIRCLE',
  'MA': 'CROSS',
  'SL': 'UNAME',
  'SQ': 'SQUARE',
  'TR': 'TRIANGLE'
};


// type of prop values
function arrange_by_type() {
    let ptype={};
    // eslint-disable-next-line no-unused-vars
    Object.entries(PROPS).forEach(([key, value]) => {
        //console.log("3",`${key}: ${value}`)); 
        const t = value[1];
        if (!(t in ptype)) ptype[t] = []; // !ptype.[t]
        ptype[t].push(key)
    });
    // iterate through key-value gracefully
    // eslint-disable-next-line no-unused-vars
    //for (const [key, value] of Object.entries(PROPS)) {
        //console.log(`${key} ${value}`); // "a 5", "b 7", "c 9"
    //}
    //console.log(ptype)
    return ptype;
}

// 
// eslint-disable-next-line no-unused-vars
const PROPS_TYPES = arrange_by_type()
export const testProps = PROPS_TYPES;

 // eslint-disable-next-line no-unused-vars
class SGFNode {
    value = ''; //: String;
    props = [];
    variations = []; // children SGFNode
    level = 0;
    seq = 0;
    root = false;
}

/* eslint-disable no-unused-vars */
export function parseSGF (stream) {
    let sgf = {} // object to retrun
    let errors = []; // stack errors to retrun

    let nNode = 0; // current node position in current tree
    let pNode = 0; // position in parent tree (redundant with pTree[last])
    let trees = {}; // array of root nodes; ID = level + position in parent
    let nodes = []; // list of SGFNode in the current tree (variation)
    let currentNode = null;
    let currentTreeLevel = -1;
    let currentTreeID = 0;
    let inPropIdent = false;
    let inPropValue = false;
    let propIdent = "";
    let propValue = ""; // TODO array to handle setup
    let escape = false;
    let inPointList = false; // AB [we][wq][we]
    let inSetup = false;
    let setup = {AW:[], AB:[], AE:[], PL:null}
    let lastIdent = ""; // for setup, last non blank
    let rootNode = null; // TODO isolate root node
    
    let i = 0;
    let c = '';

    if (_DEBUG_) console.log("Parse SGF",stream.length,"\n",stream)

    function error(message) {
        let before = stream.substring(Math.max(0,i-5),Math.max(i-1,0));
        let after = stream.substring(i+1,Math.min(i+5,stream.length-1));
        console.log("Error Parsing SGF",i,before+"==="+c+"==="+after,message);
        console.log("  --- state:",inPropIdent,inPropValue,propIdent,propValue,escape,inPointList);
        errors.push({i,excerpt:before+"==="+c+"==="+after,message})
    }

    // TODO manage setup separately like properties with muktiple values
    function processProp() {
        if (inPointList) {
            if (inSetup) setup[lastIdent].push(propValue)
            currentNode.props.at(-1).push(propValue);
        } else {
            if (currentNode == null) {
                currentNode = {};
                currentNode.props=[];
            }   
            currentNode.props.push([propIdent,propValue]);
        }
    }

    function processTree() {
        if (inPropValue) error("Closing Tree while still within Prop Value (closing ] not reached)")
        else {
                if (_DEBUG_) console.log("trees:",currentTreeLevel, currentTreeID,nodes,trees);
                trees[[currentTreeLevel,currentTreeID]].variations = nodes;
                nodes = [];
                rootNode = true;
        }
    }

    function processNode() {
        if (inPropValue) 
            error("Closing Node while still within Prop Value (closing ] not reached)");
        if (inPointList) { // non blank ID: finish any previous setup AB, AW, AE
            currentNode.props.at(-1).push([propIdent,propValue]);
            inPointList = false;
        }
        if (currentNode != null) {
            if (rootNode) {
                nodes.push(currentNode);
                //console.log("prop:",propIdent, propValue);
            } else {
                //currentNode[]
                nodes.push(currentNode);
            }
            currentNode = null;
            nNode ++; 
    }
    }

    function processChar(c) {
        if (inPropIdent) {
            if (inPointList) { // non blank ID: finish any previous setup AB, AW, AE
                processProp();
                inPointList = false;
                inSetup=false;
            }
            if (c >= 'A' && c <= 'Z') propIdent += c; // drops whatever is not A-Z
        }
        else if (inPropValue) {
            propValue += c;
        }
        else {
            // ((c <= 32 && c >= 0) || c == 127)
            if (! /[ \f\n\r\t\v]/.test(c)) {
                console.log("ERROR char not whitespace: ",c)
            }
        }
    }

    // TODO should be arrays
    let propPtrInTree=-1 // prop pointer in tree 
    let nodePtrInTree=-1 // node pointer in tree 

    for (i = 0; i < stream.length; i++) {
    //for (let c of stream) {
        c = stream[i];
        if (c == '\\') {
            escape = true;
            if (_DEBUG_) console.log("--- escaped ---")
        } else {
            if (!escape) {
                switch (c) {
                case '(': 
                    if (inPropValue) processChar(c)
                    else {
                        processNode();
                        inPropValue = false; 
                        inPropIdent = false;
                        currentTreeLevel++; 
                        currentTreeID++;
                        currentTreeID = 0;
                        trees[[currentTreeLevel,currentTreeID]] = {};  // future root node 
                    }
                    break;
                case ')': 
                    //if (inPropIdent) error("End of tree in prop ident");
                    if (inPropValue) processChar(c)
                    else {   
                        if (currentTreeLevel < 0) error("Closing tree not opened");
                        else {
                            processNode();
                            processTree();
                            //console.log(tree[currentTreeLevel])
                            currentTreeLevel--;
//                            currentTreeID = 0;
                        }
                    }
                    break;
                case '[': 
                    if (inPropValue) processChar(c)
                    else {   
                        if (!inPropIdent) error("End of prop ident but not in prop ident");
                        inPropValue = true; 
                        inPropIdent = false;
                        //console.log(propIdent);
                        if (propIdent && (PROPS[propIdent] == undefined)) error("Unknown property "+propIdent)
                        else {
                            if (propIdent && (PROPS[propIdent][1] == POINT_LIST)) { 
                                if (propIdent && (SETUP_PROPS.includes(propIdent))) { // ["AB","AW","AE"].includes(propIdent)) {
                                    inSetup = true;
                                }
                                console.log("List prop",propIdent)
                                inPointList = true;
                            }
                        }
                        if (propIdent!="") lastIdent=propIdent;
                    }
                    break;
                case ']': 
                    if (!inPropValue) error("End of prop value but not in prop value");
                    processProp() // setup wait until next char to trigger process
                    propIdent = ""
                    propValue = ""
                    inPropValue = false;
                    inPropIdent = true; // waitig for next prop
                    break;
                case ';': 
                    //if (inPropIdent) error("Start of node in prop ident");
                    if (inPropValue) processChar(c)
                    else {                    
                        processNode();
                        if (rootNode == null) rootNode = true;
                        else rootNode = false;
                        if (currentNode) currentNode.rootNode = rootNode;
                        inPropIdent=true;
                        inPropValue = false; // just in case of error
                    }
                    break;
                default: processChar(c)
                }   
            } else { // loop after after escape
                escape = false;
                if (_DEBUG_) console.log("--- un escaped ---",inPropValue,c)
                if (inPropValue) {
                    propValue += c;
                } else{
                    error("escaped character outside prop value: "+c)
                }
            }
        }
    }
    // if still open: close node or tree
    // game set: array of root nodes, one for each game
    return {errors, trees}
}

// gameRecord (root{},setup{B,W,E,P},variations[],other{})
/* eslint-enable no-unused-vars */
export function printGame(gameRecord) {
    let ret = "";
    Object.entries(gameRecord.root).forEach(([key, value]) => {
        ret += `${key}: ${value}\n`;
        //console.log("3",`${key}: ${value}`)); 
    });
    ret += "\n";
    // eslint-disable-next-line no-unused-vars
    Object.entries(gameRecord.variations[0]).forEach(([key, value]) => {

    });
    // eslint-disable-next-line no-unused-vars
    let currentNode = gameRecord.start;

    //for (const [key, value] of Object.entries(PROPS)) {
    return ret;
}

// eslint-disable-next-line no-unused-vars
function walk(tree) {
  var indent = 1;
  function innerWalk(tree) {
    tree.forEach(function(node) {
      console.log('--' + Array(indent).join('--'), node.key);
      if(node.children) {
        indent ++;
        innerWalk(node.children);
      }
      if(tree.indexOf(node) === tree.length - 1) {
        indent--;
      }
    })
  }
  innerWalk(tree);
}

// eslint-disable-next-line no-unused-vars
function preOrder(node) { 
    console.log(node.val) 
    node.left && this.preOrder(node.left) 
    node.right && this.preOrder(node.right) 
  }
  
 // node.preOrder(node) 
